@@ -72,6 +72,10 @@ use std::{
 pub mod data_structures;
 pub use data_structures::*;
 
+/// R1CS constraints for polynomial constraints.
+mod constraints;
+pub use constraints::*;
+
 /// Errors pertaining to query sets.
 pub mod error;
 pub use error::*;
@@ -157,8 +161,12 @@ pub trait PolynomialCommitment<F: Field>: Sized + Clone + Debug {
     type CommitterKey: PCCommitterKey + Clone;
     /// The verifier key for the scheme; used to check an evaluation proof.
     type VerifierKey: PCVerifierKey + Clone;
+    /// The prepared verifier key for the scheme; used to check an evaluation proof.
+    type PreparedVerifierKey: PCPreparedVerifierKey<Self::VerifierKey> + Clone;
     /// The commitment to a polynomial.
     type Commitment: PCCommitment + Clone;
+    /// The prepared commitment to a polynomial.
+    type PreparedCommitment: PCPreparedCommitment<Self::Commitment>;
     /// The commitment randomness.
     type Randomness: PCRandomness + Clone;
     /// The evaluation proof for a single point.
@@ -180,7 +188,7 @@ pub trait PolynomialCommitment<F: Field>: Sized + Clone + Debug {
     /// Specializes the public parameters for polynomials up to the given `supported_degree`
     /// and for enforcing degree bounds in the range `1..=supported_degree`.
     fn trim(
-        pp: &Self::UniversalParams,
+        parameters: &Self::UniversalParams,
         supported_degree: usize,
         supported_hiding_bound: usize,
         enforced_degree_bounds: Option<&[usize]>,
@@ -307,7 +315,7 @@ pub trait PolynomialCommitment<F: Field>: Sized + Clone + Debug {
     /// committed in `labeled_commitments`.
     fn batch_check<'a, R: RngCore>(
         vk: &Self::VerifierKey,
-        commitments: impl Iterator<Item = LabeledCommitment<Self::Commitment>>,
+        commitments: impl IntoIterator<Item = &'a LabeledCommitment<Self::Commitment>>,
         query_set: &QuerySet<F>,
         evaluations: &Evaluations<F>,
         proof: &Self::BatchProof,
@@ -317,7 +325,7 @@ pub trait PolynomialCommitment<F: Field>: Sized + Clone + Debug {
     where
         Self::Commitment: 'a,
     {
-        let commitments: BTreeMap<_, _> = commitments.map(|c| (c.label().to_owned(), c)).collect();
+        let commitments: BTreeMap<_, _> = commitments.into_iter().map(|c| (c.label(), c)).collect();
         let mut query_to_labels_map = BTreeMap::new();
         for (label, point) in query_set.iter() {
             let labels = query_to_labels_map.entry(point).or_insert_with(BTreeSet::new);
@@ -398,7 +406,7 @@ pub trait PolynomialCommitment<F: Field>: Sized + Clone + Debug {
     fn check_combinations<'a, R: RngCore>(
         vk: &Self::VerifierKey,
         linear_combinations: impl IntoIterator<Item = &'a LinearCombination<F>>,
-        commitments: impl Iterator<Item = LabeledCommitment<Self::Commitment>>,
+        commitments: impl IntoIterator<Item = &'a LabeledCommitment<Self::Commitment>>,
         eqn_query_set: &QuerySet<F>,
         eqn_evaluations: &Evaluations<F>,
         proof: &BatchLCProof<F, Self>,
@@ -461,6 +469,36 @@ pub trait PolynomialCommitment<F: Field>: Sized + Clone + Debug {
 
         Ok(true)
     }
+
+    /// On input a list of polynomials, linear combinations of those polynomials,
+    /// and a query set, `open_combination` outputs a proof of evaluation of
+    /// the combinations at the points in the query set.
+    fn open_combinations_individual_opening_challenges<'a>(
+        ck: &Self::CommitterKey,
+        linear_combinations: impl IntoIterator<Item = &'a LinearCombination<F>>,
+        polynomials: impl IntoIterator<Item = &'a LabeledPolynomial<F>>,
+        commitments: impl IntoIterator<Item = &'a LabeledCommitment<Self::Commitment>>,
+        query_set: &QuerySet<F>,
+        opening_challenges: &dyn Fn(u64) -> F,
+        rands: impl IntoIterator<Item = &'a Self::Randomness>,
+    ) -> Result<BatchLCProof<F, Self>, Self::Error>
+    where
+        Self::Randomness: 'a,
+        Self::Commitment: 'a;
+
+    /// Check combinations with individual challenges.
+    fn check_combinations_individual_opening_challenges<'a, R: RngCore>(
+        vk: &Self::VerifierKey,
+        linear_combinations: impl IntoIterator<Item = &'a LinearCombination<F>>,
+        commitments: impl IntoIterator<Item = &'a LabeledCommitment<Self::Commitment>>,
+        eqn_query_set: &QuerySet<F>,
+        eqn_evaluations: &Evaluations<F>,
+        proof: &BatchLCProof<F, Self>,
+        opening_challenges: &dyn Fn(u64) -> F,
+        rng: &mut R,
+    ) -> Result<bool, Self::Error>
+    where
+        Self::Commitment: 'a;
 }
 
 /// Evaluate the given polynomials at `query_set`.
@@ -515,6 +553,17 @@ pub mod tests {
         num_equations: Option<usize>,
     }
 
+    pub struct TestComponents<F: Field, PC: PolynomialCommitment<F>> {
+        pub verification_key: PC::VerifierKey,
+        pub commitments: Vec<LabeledCommitment<PC::Commitment>>,
+        pub query_set: QuerySet<'static, F>,
+        pub evaluations: Evaluations<'static, F>,
+        pub batch_lc_proof: Option<BatchLCProof<F, PC>>,
+        pub batch_proof: Option<PC::BatchProof>,
+        pub opening_challenge: F,
+        pub randomness: Vec<PC::Randomness>,
+    }
+
     pub fn bad_degree_bound_test<F, PC>() -> Result<(), PC::Error>
     where
         F: Field,
@@ -541,12 +590,7 @@ pub mod tests {
                 let hiding_bound = Some(1);
                 degree_bounds.push(degree_bound);
 
-                polynomials.push(LabeledPolynomial::new_owned(
-                    label,
-                    poly,
-                    Some(degree_bound),
-                    hiding_bound,
-                ))
+                polynomials.push(LabeledPolynomial::new(label, poly, Some(degree_bound), hiding_bound))
             }
 
             println!("supported degree: {:?}", supported_degree);
@@ -575,21 +619,13 @@ pub mod tests {
                 &rands,
                 Some(rng),
             )?;
-            let result = PC::batch_check(
-                &vk,
-                comms.into_iter(),
-                &query_set,
-                &values,
-                &proof,
-                opening_challenge,
-                rng,
-            )?;
+            let result = PC::batch_check(&vk, &comms, &query_set, &values, &proof, opening_challenge, rng)?;
             assert!(result, "proof was incorrect, Query set: {:#?}", query_set);
         }
         Ok(())
     }
 
-    fn test_template<F, PC>(info: TestInfo) -> Result<(), PC::Error>
+    fn test_template<F, PC>(info: TestInfo) -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -604,6 +640,8 @@ pub mod tests {
             ..
         } = info;
 
+        let mut test_components = Vec::new();
+
         let rng = &mut test_rng();
         let max_degree = max_degree.unwrap_or_else(|| rand::distributions::Uniform::from(2..=64).sample(rng));
         let pp = PC::setup(max_degree, rng)?;
@@ -642,7 +680,7 @@ pub mod tests {
                 };
                 println!("Hiding bound: {:?}", hiding_bound);
 
-                polynomials.push(LabeledPolynomial::new_owned(label, poly, degree_bound, hiding_bound))
+                polynomials.push(LabeledPolynomial::new(label, poly, degree_bound, hiding_bound))
             }
             let supported_hiding_bound = polynomials
                 .iter()
@@ -681,15 +719,7 @@ pub mod tests {
                 &rands,
                 Some(rng),
             )?;
-            let result = PC::batch_check(
-                &vk,
-                comms.into_iter(),
-                &query_set,
-                &values,
-                &proof,
-                opening_challenge,
-                rng,
-            )?;
+            let result = PC::batch_check(&vk, &comms, &query_set, &values, &proof, opening_challenge, rng)?;
             if !result {
                 println!(
                     "Failed with {} polynomials, num_points_in_query_set: {:?}",
@@ -701,11 +731,22 @@ pub mod tests {
                 }
             }
             assert!(result, "proof was incorrect, Query set: {:#?}", query_set);
+
+            test_components.push(TestComponents {
+                verification_key: vk,
+                commitments: comms,
+                query_set,
+                evaluations: values,
+                batch_lc_proof: None,
+                batch_proof: Some(proof),
+                opening_challenge,
+                randomness: rands,
+            });
         }
-        Ok(())
+        Ok(test_components)
     }
 
-    fn equation_test_template<F, PC>(info: TestInfo) -> Result<(), PC::Error>
+    fn equation_test_template<F, PC>(info: TestInfo) -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -720,6 +761,8 @@ pub mod tests {
             num_equations,
         } = info;
 
+        let mut test_components = Vec::new();
+
         let rng = &mut test_rng();
         let max_degree = max_degree.unwrap_or_else(|| rand::distributions::Uniform::from(2..=64).sample(rng));
         let pp = PC::setup(max_degree, rng)?;
@@ -762,7 +805,7 @@ pub mod tests {
                 };
                 println!("Hiding bound: {:?}", hiding_bound);
 
-                polynomials.push(LabeledPolynomial::new_owned(label, poly, degree_bound, hiding_bound))
+                polynomials.push(LabeledPolynomial::new(label, poly, degree_bound, hiding_bound))
             }
             let supported_hiding_bound = polynomials
                 .iter()
@@ -839,7 +882,7 @@ pub mod tests {
             let result = PC::check_combinations(
                 &vk,
                 &linear_combinations,
-                comms.into_iter(),
+                &comms,
                 &query_set,
                 &values,
                 &proof,
@@ -857,11 +900,22 @@ pub mod tests {
                 }
             }
             assert!(result, "proof was incorrect, equations: {:#?}", linear_combinations);
+
+            test_components.push(TestComponents {
+                verification_key: vk,
+                commitments: comms,
+                query_set,
+                evaluations: values,
+                batch_lc_proof: Some(proof),
+                batch_proof: None,
+                opening_challenge,
+                randomness: rands,
+            });
         }
-        Ok(())
+        Ok(test_components)
     }
 
-    pub fn single_poly_test<F, PC>() -> Result<(), PC::Error>
+    pub fn single_poly_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -878,7 +932,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn linear_poly_degree_bound_test<F, PC>() -> Result<(), PC::Error>
+    pub fn linear_poly_degree_bound_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -895,7 +949,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn single_poly_degree_bound_test<F, PC>() -> Result<(), PC::Error>
+    pub fn single_poly_degree_bound_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -912,7 +966,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn quadratic_poly_degree_bound_multiple_queries_test<F, PC>() -> Result<(), PC::Error>
+    pub fn quadratic_poly_degree_bound_multiple_queries_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -929,7 +983,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn single_poly_degree_bound_multiple_queries_test<F, PC>() -> Result<(), PC::Error>
+    pub fn single_poly_degree_bound_multiple_queries_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -946,7 +1000,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn two_polys_degree_bound_single_query_test<F, PC>() -> Result<(), PC::Error>
+    pub fn two_polys_degree_bound_single_query_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -963,7 +1017,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn full_end_to_end_test<F, PC>() -> Result<(), PC::Error>
+    pub fn full_end_to_end_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -980,7 +1034,7 @@ pub mod tests {
         test_template::<F, PC>(info)
     }
 
-    pub fn full_end_to_end_equation_test<F, PC>() -> Result<(), PC::Error>
+    pub fn full_end_to_end_equation_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -997,7 +1051,7 @@ pub mod tests {
         equation_test_template::<F, PC>(info)
     }
 
-    pub fn single_equation_test<F, PC>() -> Result<(), PC::Error>
+    pub fn single_equation_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -1014,7 +1068,7 @@ pub mod tests {
         equation_test_template::<F, PC>(info)
     }
 
-    pub fn two_equation_test<F, PC>() -> Result<(), PC::Error>
+    pub fn two_equation_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
@@ -1031,7 +1085,7 @@ pub mod tests {
         equation_test_template::<F, PC>(info)
     }
 
-    pub fn two_equation_degree_bound_test<F, PC>() -> Result<(), PC::Error>
+    pub fn two_equation_degree_bound_test<F, PC>() -> Result<Vec<TestComponents<F, PC>>, PC::Error>
     where
         F: Field,
         PC: PolynomialCommitment<F>,
