@@ -14,8 +14,18 @@
 // You should have received a copy of the GNU General Public License
 // along with the snarkVM library. If not, see <https://www.gnu.org/licenses/>.
 
-use crate::{impl_bytes, kzg10, BTreeMap, PCCommitterKey, PCVerifierKey, Vec};
-use snarkvm_curves::traits::{PairingCurve, PairingEngine};
+use crate::{
+    impl_bytes,
+    kzg10,
+    BTreeMap,
+    PCCommitterKey,
+    PCPreparedCommitment,
+    PCPreparedVerifierKey,
+    PCVerifierKey,
+    Vec,
+};
+use snarkvm_curves::traits::{PairingCurve, PairingEngine, ProjectiveCurve};
+use snarkvm_fields::{ConstraintFieldError, ToConstraintField};
 use snarkvm_utilities::{
     bytes::{FromBytes, ToBytes},
     error,
@@ -32,6 +42,23 @@ pub type Randomness<E> = kzg10::Randomness<E>;
 /// `Commitment` is the commitment for the KZG10 scheme.
 pub type Commitment<E> = kzg10::Commitment<E>;
 
+/// `PreparedCommitment` is the prepared commitment for the KZG10 scheme.
+pub type PreparedCommitment<E> = kzg10::PreparedCommitment<E>;
+
+impl<E: PairingEngine> PCPreparedCommitment<Commitment<E>> for PreparedCommitment<E> {
+    /// prepare `PreparedCommitment` from `Commitment`
+    fn prepare(comm: &Commitment<E>) -> Self {
+        let mut prepared_comm = Vec::<E::G1Affine>::new();
+        let mut cur = E::G1Projective::from(comm.0.clone());
+        for _ in 0..128 {
+            prepared_comm.push(cur.clone().into());
+            cur.double_in_place();
+        }
+
+        Self { 0: prepared_comm }
+    }
+}
+
 /// `ComitterKey` is used to commit to, and create evaluation proofs for, a given
 /// polynomial.
 #[derive(Derivative)]
@@ -164,6 +191,40 @@ impl<E: PairingEngine> PCVerifierKey for VerifierKey<E> {
     }
 }
 
+impl<E: PairingEngine> ToConstraintField<E::Fq> for VerifierKey<E>
+where
+    E::G1Affine: ToConstraintField<E::Fq>,
+    E::G2Affine: ToConstraintField<E::Fq>,
+{
+    fn to_field_elements(&self) -> Result<Vec<E::Fq>, ConstraintFieldError> {
+        let mut res = Vec::new();
+        res.extend_from_slice(&self.g.to_field_elements()?);
+        res.extend_from_slice(&self.gamma_g.to_field_elements()?);
+        res.extend_from_slice(&self.h.to_field_elements()?);
+        res.extend_from_slice(&self.beta_h.to_field_elements()?);
+
+        if let Some(degree_bounds_and_prepared_neg_powers_of_h) = &self.degree_bounds_and_prepared_neg_powers_of_h {
+            for (d, _prepared_neg_powers_of_h) in degree_bounds_and_prepared_neg_powers_of_h.iter() {
+                let d_elem: E::Fq = (*d as u64).into();
+
+                res.push(d_elem);
+            }
+        }
+
+        Ok(res)
+    }
+}
+
+/// Nothing to do to prepare this verifier key (for now).
+pub type PreparedVerifierKey<E> = VerifierKey<E>;
+
+impl<E: PairingEngine> PCPreparedVerifierKey<VerifierKey<E>> for PreparedVerifierKey<E> {
+    /// prepare `PreparedVerifierKey` from `VerifierKey`
+    fn prepare(vk: &VerifierKey<E>) -> Self {
+        vk.clone()
+    }
+}
+
 /// Evaluation proof at a query set.
 #[derive(Derivative)]
 #[derivative(
